@@ -39,24 +39,19 @@ __author__ =    "Patrick Lehmann"
 __email__ =     "Paebbels@gmail.com"
 __copyright__ = "2016-2022, Patrick Lehmann"
 __license__ =   "Apache License, Version 2.0"
-__version__ =   "0.17.1"
+__version__ =   "0.20.0"
 
 
-from enum     import IntEnum, unique, Enum
-from typing   import List, Iterable, Union, Optional as Nullable, Dict, cast, Tuple
+from enum            import IntEnum, unique, Enum, Flag, auto
+
+from pyTooling.Graph import Vertex
+from typing          import List, Iterable, Union, Optional as Nullable, Dict, cast, Tuple, Any
 
 from pyTooling.Decorators import export
 
 
 SimpleOrAttribute =     Union['SimpleName',    'AttributeName']
 
-LibraryOrSymbol =       Union['Library',       'LibrarySymbol']
-EntityOrSymbol =        Union['Entity',        'EntitySymbol']
-ArchitectureOrSymbol =  Union['Architecture',  'ArchitectureSymbol']
-PackageOrSymbol =       Union['Package',       'PackageSymbol']
-ConfigurationOrSymbol = Union['Configuration', 'ConfigurationSymbol']
-ContextOrSymbol =       Union['Context',       'ContextSymbol']
-
 SubtypeOrSymbol =       Union['Subtype',       'SubtypeSymbol']
 
 ConstantOrSymbol =      Union['Constant',      'ConstantSymbol']
@@ -388,41 +383,46 @@ class EntityClass(Enum):
 
 
 @export
-class PossibleReference(IntEnum):
+class PossibleReference(Flag):
 	Unknown =         0
-	Library =         2**0
-	Entity =          2**1
-	Architecture =    2**2
-	Component =       2**3
-	Package =         2**4
-	Configuration =   2**5
-	Context =         2**6
-	Type =            2**7
-	Subtype =         2**8
-	ScalarType =      2**9
-	ArrayType =       2**10
-	RecordType =      2**11
-	AccessType =      2**12
-	ProtectedType =   2**13
-	FileType =        2**14
-#	Alias =           2**14   # TODO: Is this needed?
-	Attribute =       2**15
-	TypeAttribute =   2**16
-	ValueAttribute =  2**17
-	SignalAttribute = 2**18
-	RangeAttribute =  2**19
-	ViewAttribute =   2**20
-	Constant =        2**16
-	Variable =        2**17
-	Signal =          2**18
-	File =            2**19
-	Object =          2**20   # TODO: Is this needed?
-	EnumLiteral =     2**21
-	Procedure =       2**22
-	Function =        2**23
-	Label =           2**24
-	View =            2**25
-	SimpleNameInExpression = Constant + Variable + Signal + ScalarType + EnumLiteral + Function
+	Library =         auto()
+	Entity =          auto()
+	Architecture =    auto()
+	Component =       auto()
+	Package =         auto()
+	Configuration =   auto()
+	Context =         auto()
+	Type =            auto()
+	Subtype =         auto()
+	ScalarType =      auto()
+	ArrayType =       auto()
+	RecordType =      auto()
+	AccessType =      auto()
+	ProtectedType =   auto()
+	FileType =        auto()
+#	Alias =           auto()   # TODO: Is this needed?
+	Attribute =       auto()
+	TypeAttribute =   auto()
+	ValueAttribute =  auto()
+	SignalAttribute = auto()
+	RangeAttribute =  auto()
+	ViewAttribute =   auto()
+	Constant =        auto()
+	Variable =        auto()
+	Signal =          auto()
+	File =            auto()
+#	Object =          auto()   # TODO: Is this needed?
+	EnumLiteral =     auto()
+	Procedure =       auto()
+	Function =        auto()
+	Label =           auto()
+	View =            auto()
+
+	AnyType = ScalarType | ArrayType | RecordType | ProtectedType | AccessType | FileType | Subtype
+	Object = Constant | Variable | Signal | File
+	SubProgram = Procedure | Function
+	PackageMember = AnyType | Object | SubProgram | Component
+	SimpleNameInExpression = Constant | Variable | Signal | ScalarType | EnumLiteral | Function
 
 
 @export
@@ -455,11 +455,12 @@ class NamedEntityMixin:
 	"""
 	A ``NamedEntityMixin`` is a mixin class for all VHDL entities that have identifiers.
 
-	A protected variable :attr:`_identifier` is available to derived classes as well as a readonly property
-	:attr:`Identifier` for public access.
+	Protected variables :attr:`_identifier` and :attr:`_normalizedIdentifier` are available to derived classes as well as
+	two readonly properties :attr:`Identifier` and :attr:`NormalizedIdentifier` for public access.
 	"""
 
-	_identifier: str  #: The identifier of a model entity.
+	_identifier: str            #: The identifier of a model entity.
+	_normalizedIdentifier: str  #: The normalized (lower case) identifier of a model entity.
 
 	def __init__(self, identifier: str):
 		"""
@@ -468,6 +469,7 @@ class NamedEntityMixin:
 		:param identifier: Identifier (name) of the model entity.
 		"""
 		self._identifier = identifier
+		self._normalizedIdentifier = identifier.lower()
 
 	@property
 	def Identifier(self) -> str:
@@ -478,18 +480,28 @@ class NamedEntityMixin:
 		"""
 		return self._identifier
 
+	@property
+	def NormalizedIdentifier(self) -> str:
+		"""
+		Returns a model entity's normalized identifier (lower case name).
+
+		:returns: Normalized name of a model entity.
+		"""
+		return self._normalizedIdentifier
+
 
 @export
 class MultipleNamedEntityMixin:
 	"""
 	A ``MultipleNamedEntityMixin`` is a mixin class for all VHDL entities that declare multiple instances at once by
-	giving multiple identifiers.
+	defining multiple identifiers.
 
-	A protected variable :attr:`_identifiers` is available to derived classes as well as a readonly property
-	:attr:`Identifiers` for public access.
+	Protected variables :attr:`_identifiers` and :attr:`_normalizedIdentifiers` are available to derived classes as well
+	as two readonly properties :attr:`Identifiers` and :attr:`NormalizedIdentifiers` for public access.
 	"""
 
-	_identifiers: Tuple[str]  #: A list of identifiers.
+	_identifiers:           Tuple[str]  #: A list of identifiers.
+	_normalizedIdentifiers: Tuple[str]  #: A list of normalized (lower case) identifiers.
 
 	def __init__(self, identifiers: Iterable[str]):
 		"""
@@ -498,6 +510,7 @@ class MultipleNamedEntityMixin:
 		:param identifiers: Sequence of identifiers (names) of the model entity.
 		"""
 		self._identifiers = tuple(identifiers)
+		self._normalizedIdentifiers = tuple([identifier.lower() for identifier in identifiers])
 
 	@property
 	def Identifiers(self) -> Tuple[str]:
@@ -508,16 +521,26 @@ class MultipleNamedEntityMixin:
 		"""
 		return self._identifiers
 
+	@property
+	def NormalizedIdentifiers(self) -> Tuple[str]:
+		"""
+		Returns a model entity's tuple of normalized identifiers (lower case names).
+
+		:returns: Tuple of normalized identifiers.
+		"""
+		return self._normalizedIdentifiers
+
 
 @export
 class LabeledEntityMixin:
 	"""
 	A ``LabeledEntityMixin`` is a mixin class for all VHDL entities that can have labels.
 
-	A protected variable :attr:`_label` is available to derived classes as well as a readonly property :attr:`Label` for
-	public access.
+	protected variables :attr:`_label` and :attr:`_normalizedLabel` are available to derived classes as well as two
+	readonly properties :attr:`Label` and :attr:`NormalizedLabel` for public access.
 	"""
-	_label: Nullable[str]  #: The label of a model entity.
+	_label:           Nullable[str]  #: The label of a model entity.
+	_normalizedLabel: Nullable[str]  #: The normalized (lower case) label of a model entity.
 
 	def __init__(self, label: Nullable[str]):
 		"""
@@ -526,6 +549,7 @@ class LabeledEntityMixin:
 		:param label: Label of the model entity.
 		"""
 		self._label = label
+		self._normalizedLabel = label.lower() if label is not None else None
 
 	@property
 	def Label(self) -> Nullable[str]:
@@ -536,6 +560,15 @@ class LabeledEntityMixin:
 		"""
 		return self._label
 
+	@property
+	def NormalizedLabel(self) -> Nullable[str]:
+		"""
+		Returns a model entity's normalized (lower case) label.
+
+		:returns: Normalized label of a model entity.
+		"""
+		return self._normalizedLabel
+
 
 @export
 class DocumentedEntityMixin:
@@ -567,7 +600,124 @@ class DocumentedEntityMixin:
 
 
 @export
-class DesignUnitWithContextMixin:
+class Name(ModelEntity):
+	"""``Name`` is the base-class for all *names* in the VHDL language model."""
+
+	_identifier: str
+	_normalizedIdentifier: str
+	_root: Nullable['Name']     # TODO: seams to be unused. There is no reverse linking
+	_prefix: Nullable['Name']
+
+	def __init__(self, identifier: str, prefix: 'Name' = None):
+		super().__init__()
+
+		self._identifier = identifier
+		self._normalizedIdentifier = identifier.lower()
+
+		if prefix is None:
+			self._prefix = self
+			self._root = None
+		else:
+			self._prefix = prefix
+			self._root = prefix._root
+
+	@property
+	def Identifier(self) -> str:
+		return self._identifier
+
+	@property
+	def NormalizedIdentifier(self) -> str:
+		return self._normalizedIdentifier
+
+	@property
+	def Root(self) -> 'Name':
+		return self._root
+
+	@property
+	def Prefix(self) -> Nullable['Name']:
+		return self._prefix
+
+	@property
+	def Has_Prefix(self) -> bool:
+		return self._prefix is not None
+
+
+@export
+class Symbol(ModelEntity):
+	_symbolName: Name
+	_possibleReferences: PossibleReference
+	_reference: Any
+
+	def __init__(self, symbolName: Name, possibleReferences: PossibleReference):
+		super().__init__()
+
+		self._symbolName = symbolName
+		self._possibleReferences = possibleReferences
+		self._reference = None
+
+	@property
+	def SymbolName(self) -> Name:
+		return self._symbolName
+
+
+@export
+class NewSymbol:
+	_possibleReferences: PossibleReference
+	_reference: Any
+
+	def __init__(self, possibleReferences: PossibleReference):
+		self._possibleReferences = possibleReferences
+		self._reference = None
+
+	@property
+	def Reference(self) -> Any:
+		return self._reference
+
+	@property
+	def IsResolved(self) -> bool:
+		return self._reference is not None
+
+	def __bool__(self) -> bool:
+		return self._reference is not None
+
+	def __str__(self) -> str:
+		if self._reference is not None:
+			return str(self._reference)
+		return str(self._symbolName)
+
+
+@export
+class Reference(ModelEntity):
+	_symbols:       List[Symbol]
+
+	def __init__(self, symbols: Iterable[Symbol]):
+		super().__init__()
+
+		self._symbols = [s for s in symbols]
+
+	@property
+	def Symbols(self) -> List[Symbol]:
+		return self._symbols
+
+
+@export
+class LibraryClause(Reference):
+	pass
+
+
+@export
+class UseClause(Reference):
+	pass
+
+
+# TODO: rename to ContextClause?
+@export
+class ContextReference(Reference):
+	pass
+
+
+@export
+class DesignUnitWithContextMixin: #(metaclass=ExtendedType, useSlots=True):
 	_contextItems:      List['ContextUnion']      #: List of all context items (library, use and context clauses).
 	_libraryReferences: List['LibraryClause']     #: List of library clauses.
 	_packageReferences: List['UseClause']         #: List of use clauses.
@@ -579,9 +729,9 @@ class DesignUnitWithContextMixin:
 
 		:param contextItems: A sequence of library, use or context clauses.
 		"""
-		from pyVHDLModel.SyntaxModel import LibraryClause, UseClause, ContextReference
 
 		self._contextItems = []
+		# TODO: move to DesignUnit?
 		self._libraryReferences = []
 		self._packageReferences = []
 		self._contextReferences = []
@@ -634,10 +784,57 @@ class DesignUnitWithContextMixin:
 		return self._packageReferences
 
 
+@export
+@unique
+class DesignUnitKind(Flag):
+	Context = auto()
+	Package = auto()
+	PackageBody = auto()
+	Entity = auto()
+	Architecture = auto()
+	Configuration = auto()
+
+	Primary = Context | Configuration | Entity | Package
+	Secondary = Architecture | PackageBody
+	WithContext = Configuration | Entity | Package | Architecture | PackageBody
+
+	All = Primary | Secondary
+
+
+@export
+@unique
+class DependencyGraphVertexKind(Flag):
+	Library = auto()
+	Context = auto()
+	Package = auto()
+	PackageBody = auto()
+	Entity = auto()
+	Architecture = auto()
+	Configuration = auto()
+
+
+@export
+@unique
+class DependencyGraphEdgeKind(Flag):
+	Library = auto()
+	Context = auto()
+	Package = auto()
+	Entity = auto()
+	# Architecture = auto()
+	Configuration = auto()
+	Component = auto()
+
+
 @export
 class DesignUnit(ModelEntity, NamedEntityMixin, DocumentedEntityMixin):
 	"""A ``DesignUnit`` is a base-class for all design units."""
 
+	_library:            'Library'
+	_dependencyVertex:    Vertex[str, 'DesignUnit', None, None]
+	_referencedLibraries: Dict[str, 'Library']
+	_referencedPackages:  Dict[str, Dict[str, 'Package']]
+	_referencedContexts:  Dict[str, 'Context']
+
 	def __init__(self, identifier: str, documentation: str = None):
 		"""
 		Initializes a design unit.
@@ -649,19 +846,45 @@ class DesignUnit(ModelEntity, NamedEntityMixin, DocumentedEntityMixin):
 		NamedEntityMixin.__init__(self, identifier)
 		DocumentedEntityMixin.__init__(self, documentation)
 
+		self._library = None
+		self._dependencyVertex = None
+		self._referencedLibraries = {}
+		self._referencedPackages = {"work": {}}  # TODO: should it be the working library name ... auto generated elsewhere already
+		self._referencedContexts = {}
+
+	@property
+	def Document(self) -> 'Document':
+		return self._parent
+
+	@Document.setter
+	def Document(self, document: 'Document') -> None:
+		self._parent = document
+
+	@property
+	def Library(self) -> 'Library':
+		return self._library
+
+	@Library.setter
+	def Library(self, library: 'Library') -> None:
+		self._library = library
+
+	@property
+	def ReferencedLibraries(self) -> Dict[str, 'Library']:
+		return self._referencedLibraries
+
+	@property
+	def ReferencedPackages(self) -> Dict[str, 'Package']:
+		return self._referencedPackages
+
+	@property
+	def ReferencedContexts(self) -> Dict[str, 'Context']:
+		return self._referencedContexts
+
 
 @export
 class PrimaryUnit(DesignUnit):
 	"""A ``PrimaryUnit`` is a base-class for all primary units."""
 
-	@property
-	def Library(self) -> 'Library':
-		return self._parent
-
-	@Library.setter
-	def Library(self, library: 'Library') -> None:
-		self._parent = library
-
 
 @export
 class SecondaryUnit(DesignUnit):
